/*
**  LinVPN 1.2
**  Copyright (C) 2001 Alex Fiori <alex@linuxbr.com>
**
**  This program is free software; you can redistribute it and/or
**  modify it under the terms of the GNU General Public License
**  as published by the Free Software Foundation; either version 2
**  of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307
**  USA, or contact Alex Fiori <alex@linuxbr.com>
*/

#include <pwd.h>
#include <grp.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>
#include <signal.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>

/* VPN stuff */
#include "vpn-base.h"

static struct {
	int is_used;
	char addr[17];
	char filename[1024];
} config[128];

#ifndef PREFIX
#error "See documentation"
#endif

#ifndef USER
#error "See documentation"
#endif

#ifndef GROUP
#error "See documentation"
#endif

static void go_background (void);
static void setpermissions (void);
static void log (const char *fmt, ...);
static int  readconf (const char *filename);
static void run (const char *path, const char *iface, const char *option);
static void check (const char *path, const char *ifname, const char *addr);


int main ()
{
	int i;
	char temp[1024], path[1024], *addr;

	signal (SIGINT,  SIG_IGN);
	signal (SIGHUP,  SIG_IGN);
	signal (SIGCHLD, SIG_IGN);

	/* read the config file */
	memset (temp, 0, sizeof (temp));
	snprintf (temp, sizeof (temp), "%s/etc/vpn-server.conf", PREFIX);
	if (!readconf (temp)) {
		fprintf (stderr, "Cannot open config file, exiting.\n");
		return 1;
	}

	/* temp directory */
	memset (path, 0, sizeof (path));
	snprintf (path, sizeof (path), "%s/.temp-vpn", PREFIX);
	check (path, NULL, NULL);

	fprintf (stderr, "LinVPN 1.2 daemon running.\n");
	fprintf (stderr, "Copyright (C) 2001 Alex Fiori <alex@linuxbr.com>\n");

	/* main loop */
	setpermissions ();
	go_background ();
	for (;;) {
		for (i = 0; i < 20; i++) {
			memset (temp, 0, sizeof (temp));
			snprintf (temp, sizeof (temp), "ppp%d", i);

			addr = vpn_ifaddr (temp);
			check (path, temp, addr);
		}

		sleep (2);
	}

	return 0;
}

static void go_background (void)
{
	/* welcome to the multitask world */
	switch (fork ()) {
		case -1:
			perror ("fork"), exit (1);
			exit (1);
		case 0:
			close (0); close (1); close (2);
			break;
		default:
			_exit (1);
	}

	return;
}

static void setpermissions (void)
{
	struct group  *grp;
	struct passwd *pwd;

	pwd = getpwnam (USER);
	grp = getgrnam (GROUP);

	if (pwd == NULL || grp == NULL) {
		fprintf (stderr, "Fatal error, exiting.\n");
		exit (1);
	}

	/* S E C U R I T Y - I hate people talking about security */
	setuid (pwd->pw_uid);
	setgid (grp->gr_gid);

	return;
}

static void log (const char *fmt, ...)
{
	FILE *fp;
	time_t t;
	va_list ap;
	struct tm *tm;
	char path[1024];

	memset (path, 0, sizeof (path));
	snprintf (path, sizeof (path), "%s/log/server.log", PREFIX);

	/* open! */
	if ((fp = fopen (path, "a")) == NULL) return;

	time (&t);
	tm = localtime (&t);
	fprintf (fp, "%d/%02d/%02d %02d:%02d:%02d ",
		tm->tm_year + 1900, tm->tm_mon, tm->tm_mday,
		tm->tm_hour, tm->tm_mon, tm->tm_sec);
	
	va_start (ap, fmt);
	vfprintf (fp, fmt, ap);
	va_end (ap);

	fflush (fp);
	fclose (fp);

	return;
}

static int readconf (const char *filename)
{
	FILE *fp;
	int i, size;
	char temp[1024];

	memset (&config, 0, sizeof (config));
	size = sizeof (config) / sizeof (config[0]);

	if ((fp = fopen (filename, "r")) == NULL) return 0;
	
	while (!feof (fp)) {
		memset (temp, 0, sizeof (temp));
		if (fgets (temp, sizeof (temp), fp) == NULL) break;
		if (*temp == '#') continue;

		/* find a free field */
		for (i = 0; i < size; i++)
			if (!config[i].is_used) break;

		/* too many rules */
		if (i == (size - 1)) return 0;

		config[i].is_used = 1;
		sscanf (temp, "%[^]:]:%s\n", 
			config[i].addr, config[i].filename);
	}

	fclose (fp);
	return 1;
}

static void run (const char *path, const char *iface, const char *option)
{
	char temp[1024];

	/* exec nothing */
	if (*path == '\0') return;

	memset (temp, 0, sizeof (temp));
	snprintf (temp, sizeof (temp), "%s %s %s", path, iface, option);

	/* heh.. I know it sucks but... */
	system (temp);

	return;
}

static void check (const char *path, const char *ifname, const char *addr)
{
	FILE *fp;
	char temp[1024];
	int i, n = 0, found = 0;
	struct dirent **namelist;
	static int csize = sizeof (config) / sizeof (config[0]);

	/* scandir (create directory if doesn't exist) */
	for (i = 0; i < 2; i++) {
		n = scandir (path, &namelist, 0, alphasort);
		if (n < 0) {
			if (ifname == NULL && addr == NULL) {
				if (mkdir (path, 0755) < 0) {
					perror ("Cannot create directory"),
					exit (1);
				}
			} else perror ("scandir"), exit (1);
		} else break;
	}

	/* starting, remove all files... */
	if (ifname == NULL && addr == NULL) {
		while (n--) {

		if (*namelist[n]->d_name == '.' ||
		    !strcmp (namelist[n]->d_name, "..")) continue;

		memset (temp, 0, sizeof (temp));
		snprintf (temp, sizeof (temp), "%s/%s",
			  path, namelist[n]->d_name);

		if (unlink (temp) < 0)
			fprintf (stderr, "Cannot remove file %s, exiting.",
					 temp), exit (1);
		}

		/* just finished! */
		return;
	}

	/* ok, lets work... */
	while (n--) {
		if (!strcmp (namelist[n]->d_name, ifname)) {
			found = 1;
			break;
		}
	}

	/* a new VPN is comming */
	memset (temp, 0, sizeof (temp));
	snprintf (temp, sizeof (temp), "%s/%s", path, ifname);
	if (!found && addr != NULL) {
		if ((fp = fopen (temp, "w")) == NULL)
			perror ("open"), exit (1);
	
		/* do I have rules? o_O */
		memset (temp, 0, sizeof (temp));
		*temp = '.';

		for (i = 0; i < csize; i++) {
			if (!config[i].is_used) continue;

			if (!strcmp (config[i].addr, addr)) {
			    memset (temp, 0, sizeof (temp));
			    snprintf (temp, sizeof (temp), config[i].filename);
			    run (temp, ifname,  "up");
			}
		}

		fprintf (fp, "%s:%s\n", addr, temp);
		fclose (fp);

		log ("interface %s (%s) up\n", ifname, addr);

	/* I found the file but I don't have that VPN anymore, unlink it! */
	} else if (found && addr == NULL) {
		char buff[1024], file[1024], my_addr[17];

		memset (temp, 0, sizeof (temp));
		snprintf (temp, sizeof (temp), "%s/%s", path, ifname);
		if ((fp = fopen (temp, "r")) == NULL) return;

		memset (buff, 0, sizeof (buff));
		fgets (buff, sizeof (buff), fp);

		memset (file, 0, sizeof (file));
		memset (my_addr, 0, sizeof (my_addr));
		sscanf (buff, "%[^]:]:%s\n", my_addr, file);
		fclose (fp);

		log ("interface %s (%s) down\n", ifname, my_addr);

		run (file, ifname, "down");
		unlink (temp);
	}

	return;
}
